package bzr4idea;
/*
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for
 * the specific language governing permissions and limitations under the License.
 *
 * Copyright 2007 Decentrix Inc
 * Copyright 2007 Aspiro AS
 * Copyright 2008 MQSoftware
 * Authors: gevession, Erlend Simonsen & Mark Scott
 *
 * This code was originally derived from the MKS & Mercurial IDEA VCS plugins
 */

import bzr4idea.commands.GitCommand;
import bzr4idea.commands.GitSimpleHandler;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.vcs.history.VcsRevisionNumber;
import com.intellij.openapi.vfs.VirtualFile;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

import java.util.Arrays;
import java.util.Date;
import java.util.StringTokenizer;

/**
 * Git revision number
 */
public class BzrRevisionNumber implements VcsRevisionNumber {
  /**
   * the name of tip revision
   */
  @NonNls public static final String TIP = "HEAD";
  /**
   * the hash from 40 zeros representing not yet created commit
   */
  public static final String NOT_COMMITED_HASH;

  static {
    char[] data = new char[40];
    Arrays.fill(data, '0');
    NOT_COMMITED_HASH = new String(data);
  }

  @NotNull private final String myRevision;
  /**
   * the revision number (40 character hashcode, tag, or reference). In some cases incomplete hashcode could be used.
   */
  /* git4idea
  @NotNull private final String myRevisionStr;
  */
  /**
   * the date when revision created
   */
  @NotNull private final Date myTimestamp;

  private static final Logger LOG = Logger.getInstance(BzrRevisionNumber.class);

  /**
   * A constructor from version. The current date is used.
   *
   * @param version the version number.
   */
  public BzrRevisionNumber(@NonNls @NotNull String version) {
    // TODO review usages
    /* git4idea
    myRevisionStr = version;
    */
    myRevision = version;
    myTimestamp = new Date();
  }

  /**
   * A consctructor from version and time
   *
   * @param version   the version number
   * @param timeStamp the time when the version has been created
   */
  public BzrRevisionNumber(@NotNull String version, @NotNull Date timeStamp) {
    myTimestamp = timeStamp;
    /* git4idea
    myRevisionStr = version;
    */
    myRevision = version;
  }

  /**
   * {@inheritDoc}
   *
   * @see #getRev()
   */
  @NotNull
  public String asString() {
    /* git4idea
    return myRevisionStr;
    */
    return myRevision.toString();
  }

  /**
   * @return revision time
   */
  @NotNull
  public Date getTimestamp() {
    return myTimestamp;
  }

  /**
   * @return revision number
   */
  @NotNull
  public String getRev() {
    /* git4idea
    return myRevisionStr;
    */
    return myRevision;
  }

  /**
   * @return the short revision number. The revision number likely unambiguously indentify local revision, however in rare cases there could be conflicts.
   */
  @NotNull
  public String getShortRev() {
    /* git4idea
    if (myRevisionStr.length() == 0) return "";
    if (myRevisionStr.length() == 40) return myRevisionStr.substring(0, 8);
    if (myRevisionStr.length() > 40)  // revision string encoded with date too
    {
      return myRevisionStr.substring(myRevisionStr.indexOf("[") + 1, 8);
    }
    return myRevisionStr;
    */
    // TODO short version of bazaar revision?
    return myRevision;
  }

  /**
   * {@inheritDoc}
   */
  public int compareTo(VcsRevisionNumber crev) {
    if (this == crev) return 0;

    if (crev instanceof BzrRevisionNumber) {
      BzrRevisionNumber crevg = (BzrRevisionNumber)crev;
      /* git4idea
      if ((crevg.myRevisionStr != null) && myRevisionStr.equals(crevg.myRevisionStr)) {
        return 0;
      }
      */
      if ((crevg.myRevision != null) && myRevision.equals(crevg.myRevision)) {
        return 0;
      }

      /* git4idea
      if ((crevg.myRevisionStr.indexOf("[") > 0) && (crevg.myTimestamp != null)) {
        return myTimestamp.compareTo(crevg.myTimestamp);
      }

      // check for parent revs
      String crevName = null;
      String revName = null;
      int crevNum = -1;
      int revNum = -1;

      if (crevg.myRevisionStr.contains("~")) {
        int tildeIdx = crevg.myRevisionStr.indexOf('~');
        crevName = crevg.myRevisionStr.substring(0, tildeIdx);
        crevNum = Integer.parseInt(crevg.myRevisionStr.substring(tildeIdx));
      }

      if (myRevisionStr.contains("~")) {
        int tildeIdx = myRevisionStr.indexOf('~');
        revName = myRevisionStr.substring(0, tildeIdx);
        revNum = Integer.parseInt(myRevisionStr.substring(tildeIdx));
      }

      if (crevName == null && revName == null) {
        return myTimestamp.compareTo(crevg.myTimestamp);
      }
      else if (crevName == null) {
        return 1;  // I am an ancestor of the compared revision
      }
      else if (revName == null) {
        return -1; // the compared revision is my ancestor
      }
      else {
        return revNum - crevNum;  // higher relative rev numbers are older ancestors
      }
      */
      return myRevision.compareTo(crevg.myRevision);
    }

    return -1;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) return true;
    if ((obj == null) || (obj.getClass() != getClass())) return false;

    BzrRevisionNumber test = (BzrRevisionNumber)obj;
    /* git4idea
    // TODO normailize revision string?
    return myRevisionStr.equals(test.myRevisionStr);
    */
    return myRevision.equals(test.myRevision);
  }

  @Override
  public int hashCode() {
    /* git4idea
    return myRevisionStr.hashCode();
    */
    return myRevision.hashCode();
  }

  /**
   * Create revision number from string
   *
   * @param rev a revision string
   * @return revision number (the same as {@link #BzrRevisionNumber(String)}
   */
  public static BzrRevisionNumber createRevision(String rev) {
    return new BzrRevisionNumber(rev);
  }

  /**
   * Create revision number from string
   *
   * @param rev a revision string
   * @param timestamp a revision date
   * @return revision number (the same as {@link #BzrRevisionNumber(String)}
   */
  public static BzrRevisionNumber createRevision(String rev, Date timestamp) {
    return new BzrRevisionNumber(rev, timestamp);
  }

  /**
   * Resolve revision number for the specified revision
   *
   * @param project a project
   * @param vcsRoot a vcs root
   * @param rev     a revision expression
   * @return a resolved revision number with correct time
   */
  public static BzrRevisionNumber resolve(Project project, VirtualFile vcsRoot, @NonNls String rev) throws VcsException {
    GitSimpleHandler h = new GitSimpleHandler(project, vcsRoot, GitCommand.REV_LIST);
    h.setSilent(true);
    h.addParameters("--timestamp", "--max-count=1", rev);
    h.endOptions();
    final String output = h.run();
    return parseRevlistOutputAsRevisionNumber(h, output);
  }

  @NotNull
  public static BzrRevisionNumber parseRevlistOutputAsRevisionNumber(@NotNull GitSimpleHandler h, @NotNull String output) {
    StringTokenizer tokenizer = new StringTokenizer(output, "\n\r \t", false);
    LOG.assertTrue(tokenizer.hasMoreTokens(), "No required tokens in the output: \n" + output);
    Date timestamp = GitUtil.parseTimestampWithNFEReport(tokenizer.nextToken(), h, output);
    return new BzrRevisionNumber(tokenizer.nextToken(), timestamp);
  }
}
