package bzr4idea;

import bzr4idea.commands.BzrImpl;
import bzr4idea.config.GitExecutableValidator;
import bzr4idea.util.Log;

/*
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for
 * the specific language governing permissions and limitations under the License.
 *
 * Copyright 2007 Decentrix Inc
 * Copyright 2007 Aspiro AS
 * Copyright 2008 MQSoftware
 * Authors: gevession, Erlend Simonsen & Mark Scott
 *
 * This code was originally derived from the MKS & Mercurial IDEA VCS plugins
 */
import com.intellij.execution.ui.ConsoleViewContentType;

import com.intellij.ide.BrowserUtil;
import com.intellij.notification.*;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.markup.TextAttributes;
import com.intellij.openapi.options.Configurable;
import com.intellij.openapi.options.ShowSettingsUtil;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.MessageType;
import com.intellij.openapi.vcs.*;
import com.intellij.openapi.vcs.changes.ChangeProvider;
import com.intellij.openapi.vcs.changes.ui.ChangesViewContentManager;
import com.intellij.openapi.vcs.checkin.CheckinEnvironment;
import com.intellij.openapi.vcs.diff.DiffProvider;
import com.intellij.openapi.vcs.diff.RevisionSelector;
import com.intellij.openapi.vcs.history.VcsHistoryProvider;
import com.intellij.openapi.vcs.history.VcsRevisionNumber;
import com.intellij.openapi.vcs.merge.MergeProvider;
import com.intellij.openapi.vcs.rollback.RollbackEnvironment;
import com.intellij.openapi.vcs.ui.VcsBalloonProblemNotifier;
import com.intellij.openapi.vcs.update.UpdateEnvironment;
import com.intellij.openapi.vfs.VirtualFile;

import bzr4idea.annotate.BzrAnnotationProvider;

import bzr4idea.changes.BzrChangeProvider;

import bzr4idea.checkin.BzrCheckinEnvironment;

import bzr4idea.config.GitVcsConfigurable;
import bzr4idea.config.BzrVcsSettings;
import bzr4idea.config.GitVersion;

import bzr4idea.diff.BzrDiffProvider;

import bzr4idea.history.BzrHistoryProvider;

import bzr4idea.i18n.GitBundle;

import bzr4idea.merge.GitMergeProvider;

import bzr4idea.rollback.BzrRollbackEnvironment;

import bzr4idea.update.GitUpdateEnvironment;

import bzr4idea.vfs.GitVFSListener;
import bzr4idea.ui.GitUIUtil;

import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.event.HyperlinkEvent;

import java.util.Date;
import java.util.Collection;


/**
 * Git VCS implementation
 */
public class BzrVcs extends AbstractVcs {
  public static final NotificationGroup NOTIFICATION_GROUP_ID = NotificationGroup.toolWindowGroup(
          "Git Messages", ChangesViewContentManager.TOOLWINDOW_ID, true);
  public static final NotificationGroup IMPORTANT_ERROR_NOTIFICATION = new NotificationGroup(
          "Git Important Messages", NotificationDisplayType.STICKY_BALLOON, true);
  public static final NotificationGroup MINOR_NOTIFICATION = new NotificationGroup(
          "Git Minor Notifications", NotificationDisplayType.BALLOON, true);

  @NonNls
  public static final String NAME = "Bazaar";

  /**
   * the logger
   */
  private static final Logger log = Logger.getInstance(BzrVcs.class.getName());
  private static final VcsKey ourKey = createKey(NAME);

    /**
     * change provider
     */
    private final ChangeProvider myChangeProvider;

    /**
     * commit support
     */
    private final CheckinEnvironment myCheckinEnvironment;

    /**
     * rollback support
     */
    private final RollbackEnvironment myRollbackEnvironment;

    /**
     * update support
     */
    private final GitUpdateEnvironment myUpdateEnvironment;

    /**
     * annotate file support
     */
    private final BzrAnnotationProvider myAnnotationProvider;

    /**
     * diff provider
     */
    private final DiffProvider myDiffProvider;

    /**
     * history provider
     */
    private final VcsHistoryProvider myHistoryProvider;

    /**
     * cached instace of vcs manager for the project
     */
    private final ProjectLevelVcsManager myVcsManager;

    /**
     * project vcs settings
     */
    private final BzrVcsSettings mySettings;

    /**
     * configuration support
     */
    private final Configurable myConfigurable;

    /**
     * selector for revisions
     */
    private final RevisionSelector myRevSelector;

    /**
     * merge provider
     */
    private final GitMergeProvider myMergeProvider;

    private final GitExecutableValidator myExecutableValidator;

    /**
     * a vfs listener
     */
    private GitVFSListener myVFSListener;

    /**
     * The currently detected git version or null.
     */
    @SuppressWarnings({"FieldAccessedSynchronizedAndUnsynchronized"
    })
    private GitVersion myVersion;

    /**
     * Checking the version lock (used to prevent infinite recusion)
     */
    private final Object myCheckingVersion = new Object();

    /**
     * The path to executable at the time of version check
     */
    private String myVersionCheckExcecutable = "";

    public BzrVcs(@NotNull Project project,
                  @NotNull final BzrChangeProvider bzrChangeProvider,
                  @NotNull final BzrCheckinEnvironment bzrCheckinEnvironment,
                  @NotNull final ProjectLevelVcsManager gitVcsManager,
                  @NotNull final BzrAnnotationProvider bzrAnnotationProvider,
                  @NotNull final BzrDiffProvider bzrDiffProvider,
                  @NotNull final BzrHistoryProvider bzrHistoryProvider,
                  @NotNull final BzrRollbackEnvironment bzrRollbackEnvironment,
                  @NotNull final BzrVcsSettings bzrSettings) {
      super(project, NAME);

      myVcsManager = gitVcsManager;
      mySettings = bzrSettings;
      myChangeProvider = bzrChangeProvider;
      myCheckinEnvironment = bzrCheckinEnvironment;
      myAnnotationProvider = bzrAnnotationProvider;
      myDiffProvider = bzrDiffProvider;
      myHistoryProvider = bzrHistoryProvider;
      myRollbackEnvironment = bzrRollbackEnvironment;
      myRevSelector = new GitRevisionSelector();
      myConfigurable = new GitVcsConfigurable(mySettings, myProject);
      myUpdateEnvironment = new GitUpdateEnvironment(myProject);
      myMergeProvider = new GitMergeProvider(myProject);
      myExecutableValidator = new GitExecutableValidator(myProject, this);
    }

  public static BzrVcs getInstance(@NotNull
    Project project) {
        return (BzrVcs) ProjectLevelVcsManager.getInstance(project)
                                              .findVcsByName(NAME);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getRevisionPattern() {
        // return the full commit hash pattern, possibly other revision formats should be supported as well
        return "revno\\:\\d+";
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @NotNull
    public CheckinEnvironment getCheckinEnvironment() {
        return myCheckinEnvironment;
    }

    /**
     * {@inheritDoc}
     */
    @NotNull
    @Override
    public MergeProvider getMergeProvider() {
        return myMergeProvider;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @NotNull
    public RollbackEnvironment getRollbackEnvironment() {
        return myRollbackEnvironment;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @NotNull
    public VcsHistoryProvider getVcsHistoryProvider() {
        return myHistoryProvider;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @NotNull
    public String getDisplayName() {
        return NAME;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @Nullable
    public UpdateEnvironment getUpdateEnvironment() {
        return myUpdateEnvironment;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @Nullable
    public UpdateEnvironment getStatusEnvironment() {
        return getUpdateEnvironment();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @NotNull
    public BzrAnnotationProvider getAnnotationProvider() {
        return myAnnotationProvider;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @NotNull
    public DiffProvider getDiffProvider() {
        return myDiffProvider;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @Nullable
    public RevisionSelector getRevisionSelector() {
        return myRevSelector;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public UpdateEnvironment getIntegrateEnvironment() {
        return getUpdateEnvironment();
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings({"deprecation"
    })
    @Override
    @Nullable
    public VcsRevisionNumber parseRevisionNumber(String revision, FilePath path) {
        if ((revision == null) || (revision.length() == 0)) {
            return null;
        }

        if (revision.length() > 40) { // date & revision-id encoded string

            String datestr = revision.substring(0, revision.indexOf("["));
            String rev = revision.substring(revision.indexOf("[") + 1, 40);
            Date d = new Date(Date.parse(datestr));

            return new BzrRevisionNumber(rev, d);
        }

        if (path != null) {
            VirtualFile root = GitUtil.getGitRoot(myProject, path);

            try {
                return BzrRevisionNumber.resolve(myProject, root, revision);
            } catch (VcsException e) {
                log.error("Unexpected problem with resolving the git revision number: ",
                    e);
            }
        }

        return new BzrRevisionNumber(revision);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings({"deprecation"
    })
    @Override
    @Nullable
    public VcsRevisionNumber parseRevisionNumber(String revision) {
        return parseRevisionNumber(revision, null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isVersionedDirectory(VirtualFile dir) {
        return dir.isDirectory() && (GitUtil.gitRootOrNull(dir) != null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void activate() {
        super.activate();

        if (myVFSListener == null) {
            myVFSListener = new GitVFSListener(myProject, this);
        }

      try {
        BzrImpl.initBazaarClient(getSettings());
      } catch (BzrException e) {
        Log.error(e, "Cannot initialize bzr client");
      }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void deactivate() {
        if (myVFSListener != null) {
            myVFSListener.dispose();
            myVFSListener = null;
        }

        super.deactivate();
    }

    /**
     * {@inheritDoc}
     */
    @NotNull
    @Override
    public synchronized Configurable getConfigurable() {
        return myConfigurable;
    }

    /**
     * {@inheritDoc}
     */
    @Nullable
    public ChangeProvider getChangeProvider() {
        return myChangeProvider;
    }

    /**
     * {@inheritDoc}
     *
     * @return additional file statuses specific to Bzr
     */
    @Override
    public FileStatus[] getProvidedStatuses() {
        return new FileStatus[] { BzrFileStatus.RENAMED };
    }

    /**
     * Show errors as popup and as messages in vcs view.
     *
     * @param list   a list of errors
     * @param action an action
     */
    public void showErrors(@NotNull Collection<VcsException> list, @NotNull String action) {
      GitUIUtil.showErrors( myProject, list, action );
    }

    /**
     * Show a plain message in vcs view
     *
     * @param message a message to show
     */
    public void showMessages(@NotNull
    String message) {
        if (message.length() == 0) {
            return;
        }

        showMessage(message,
            ConsoleViewContentType.NORMAL_OUTPUT.getAttributes());
    }

    /**
     * @return vcs settings for the current project
     */
    @NotNull
    public BzrVcsSettings getSettings() {
        return mySettings;
    }

    /**
     * Show message in the VCS view
     *
     * @param message a message to show
     * @param style   a style to use
     */
    private void showMessage(@NotNull
    String message, final TextAttributes style) {
        myVcsManager.addMessageToConsoleWindow(message, style);
    }

    /**
     * Check version and report problem
     */
    public void checkVersion() {
      final String executable = mySettings.GIT_EXECUTABLE;
      try {
        myVersion = GitVersion.identifyVersion(executable);
        if (!myVersion.isSupported()) {
          log.info("Unsupported Bazaar version: " + myVersion);
          final String SETTINGS_LINK = "settings";
          final String UPDATE_LINK = "update";
          String message = String.format("The <a href='" + SETTINGS_LINK + "'>configured</a> version of Git is not supported: %s.<br/> " +
                                         "The minimal supported version is %s. Please <a href='" + UPDATE_LINK + "'>update</a>.",
                                         myVersion, GitVersion.MIN);
          IMPORTANT_ERROR_NOTIFICATION.createNotification("Unsupported Bazaar version", message,
            NotificationType.ERROR,
            new NotificationListener.Adapter() {
              @Override
              protected void hyperlinkActivated(@NotNull Notification notification, @NotNull HyperlinkEvent e) {
                if (SETTINGS_LINK.equals(e.getDescription())) {
                  ShowSettingsUtil.getInstance().showSettingsDialog(myProject, getConfigurable().getDisplayName());
                }
                else if (UPDATE_LINK.equals(e.getDescription())) {
                  BrowserUtil.browse("http://bazaar.canonical.com");
                }
              }
            }).notify(myProject);
        }
      } catch (Exception e) {
        if (getExecutableValidator().checkExecutableAndNotifyIfNeeded()) { // check executable before notifying error
          final String reason = (e.getCause() != null ? e.getCause() : e).getMessage();
          String message = GitBundle.message("vcs.unable.to.run.git", executable, reason);
          if (!myProject.isDefault()) {
            showMessage(message, ConsoleViewContentType.SYSTEM_OUTPUT.getAttributes());
          }
          VcsBalloonProblemNotifier.showOverVersionControlView(myProject, message, MessageType.ERROR);
        }
      }
    }

    /**
     * @return the configured version of git
     */
    public GitVersion getVersion() {
        return myVersion;
    }

    /**
     * Show command line
     *
     * @param cmdLine a command line text
     */
    public void showCommandLine(final String cmdLine) {
        showMessage(cmdLine,
            ConsoleViewContentType.SYSTEM_OUTPUT.getAttributes());
    }

    /**
     * The error line
     *
     * @param line a line to show
     */
    public void showErrorMessages(final String line) {
        showMessage(line, ConsoleViewContentType.ERROR_OUTPUT.getAttributes());
    }

  public static VcsKey getKey() {
    return ourKey;
  }

  @NotNull
  public GitExecutableValidator getExecutableValidator() {
    return myExecutableValidator;
  }

}
