/*
 * Copyright 2000-2008 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package bzr4idea.history;

import bzr4idea.*;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vcs.FilePath;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.vcs.changes.Change;
import com.intellij.openapi.vcs.changes.ChangeListManager;
import com.intellij.openapi.vcs.history.VcsFileRevision;
import com.intellij.openapi.vcs.history.VcsRevisionNumber;
import com.intellij.openapi.vfs.VirtualFile;

import com.intellij.util.Consumer;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;


/**
 * History utilities for Git
 */
public class GitHistoryUtils {
  /**
   * A private constructor
   */
  private GitHistoryUtils() {
  }

  /**
   * Get current revision for the file under git
   *
   * @param project  a project
   * @param filePath a file path
   * @return a revision number or null if the file is unversioned or new
   * @throws VcsException if there is problem with running git
   */
  @Nullable
  public static VcsRevisionNumber getCurrentRevision(final Project project, FilePath filePath) throws VcsException {
    final VirtualFile root = GitUtil.getGitRoot(project, filePath);
    /* git4idea
    filePath = getLastCommitName(project, filePath);
    GitSimpleHandler h = new GitSimpleHandler(project, GitUtil.getGitRoot(project, filePath), GitHandler.LOG);
    h.setNoSSH(true);
    h.setSilent(true);
    h.addParameters("-n1", "--pretty=format:%H%n%ct%n");
    h.endOptions();
    h.addRelativePaths(filePath);
    String result = h.run();
    if (result.length() == 0) {
      return null;
    }
    String[] lines = result.split("\n");
    String revstr = lines[0];
    Date commitDate = GitUtil.parseTimestamp(lines[1]);
    return new BzrRevisionNumber(revstr, commitDate);
    */
    // TODO
    throw new VcsException("TODO: GitHistoryUtils.getCurrentRevision not implemented yet.");
  }

  /**
   * Get history for the file
   *
   * @param project the context project
   * @param path    the file path
   * @return the list of the revisions
   * @throws VcsException if there is problem with running git
   */
  public static List<VcsFileRevision> history(final Project project, final FilePath path) throws VcsException {
    final VirtualFile root = GitUtil.getGitRoot(path);
    return history(project, path, root);
  }

   /**
   * Get history for the file
   *
   * @param project the context project
   * @param path    the file path
   * @return the list of the revisions
   * @throws VcsException if there is problem with running git
   */
  public static List<VcsFileRevision> history(final Project project, FilePath path, final VirtualFile root, final String... parameters) throws VcsException {
    final List<VcsFileRevision> rc = new ArrayList<VcsFileRevision>();
    final List<VcsException> exceptions = new ArrayList<VcsException>();

    history(project, path, root, new Consumer<GitFileRevision>() {
      @Override public void consume(GitFileRevision gitFileRevision) {
        rc.add(gitFileRevision);
      }
    }, new Consumer<VcsException>() {
      @Override public void consume(VcsException e) {
        exceptions.add(e);
      }
    }, parameters);
    if (!exceptions.isEmpty()) {
      throw exceptions.get(0);
    }
    return rc;
  }

  /**
   * Retrieves the history of the file, including renames.
   * @param project
   * @param path              FilePath which history is queried.
   * @param root              Git root - optional: if this is null, then git root will be detected automatically.
   * @param consumer          This consumer is notified ({@link Consumer#consume(Object)} when new history records are retrieved.
   * @param exceptionConsumer This consumer is notified in case of error while executing git command.
   * @param parameters        Optional parameters which will be added to the git log command just before the path.
   */
  public static void history(final Project project, FilePath path, @Nullable VirtualFile root,
                             final Consumer<GitFileRevision> consumer, final Consumer<VcsException> exceptionConsumer,
                             String... parameters) {
    // adjust path using change manager
    final FilePath filePath = getLastCommitName(project, path);
    final VirtualFile finalRoot;
    try {
      finalRoot = (root == null ? GitUtil.getGitRoot(project, path) : root);
    }
    catch (VcsException e) {
      exceptionConsumer.consume(e);
      return;
    }
    // TODO
//    final GitLogParser logParser = new GitLogParser(project, GitLogParser.NameStatus.STATUS,
//                                                    HASH, COMMIT_TIME, AUTHOR_NAME, AUTHOR_EMAIL, COMMITTER_NAME, COMMITTER_EMAIL, PARENTS,
//                                                    SUBJECT, BODY, RAW_BODY, AUTHOR_TIME);
//
//    final AtomicReference<String> firstCommit = new AtomicReference<String>("HEAD");
//    final AtomicReference<String> firstCommitParent = new AtomicReference<String>("HEAD");
//    final AtomicReference<FilePath> currentPath = new AtomicReference<FilePath>(filePath);
//    final AtomicReference<GitLineHandler> logHandler = new AtomicReference<GitLineHandler>();
//    final AtomicBoolean skipFurtherOutput = new AtomicBoolean();
//
//    final Consumer<GitLogRecord> resultAdapter = new Consumer<GitLogRecord>() {
//      public void consume(GitLogRecord record) {
//        if (skipFurtherOutput.get()) {
//          return;
//        }
//        if (record == null) {
//          exceptionConsumer.consume(new VcsException("revision details are null."));
//          return;
//        }
//        record.setUsedHandler(logHandler.get());
//        final GitRevisionNumber revision = new GitRevisionNumber(record.getHash(), record.getDate());
//        firstCommit.set(record.getHash());
//        final String[] parentHashes = record.getParentsHashes();
//        if (parentHashes == null || parentHashes.length < 1) {
//          firstCommitParent.set(null);
//        }
//        else {
//          firstCommitParent.set(parentHashes[0]);
//        }
//        final String message = record.getFullMessage();
//
//        FilePath revisionPath;
//        try {
//          final List<FilePath> paths = record.getFilePaths(finalRoot);
//          if (paths.size() > 0) {
//            revisionPath = paths.get(0);
//          }
//          else {
//            // no paths are shown for merge commits, so we're using the saved path we're inspecting now
//            revisionPath = currentPath.get();
//          }
//
//          final Pair<String, String> authorPair = Pair.create(record.getAuthorName(), record.getAuthorEmail());
//          final Pair<String, String> committerPair =
//            record.getCommitterName() == null ? null : Pair.create(record.getCommitterName(), record.getCommitterEmail());
//          Collection<String> parents = parentHashes == null ? Collections.<String>emptyList() : Arrays.asList(parentHashes);
//          consumer.consume(new GitFileRevision(project, revisionPath, revision, Pair.create(authorPair, committerPair), message, null,
//                                               new Date(record.getAuthorTimeStamp()), parents));
//          List<GitLogStatusInfo> statusInfos = record.getStatusInfos();
//          if (statusInfos.isEmpty()) {
//            // can safely be empty, for example, for simple merge commits that don't change anything.
//            return;
//          }
//          if (statusInfos.get(0).getType() == GitChangeType.ADDED && !filePath.isDirectory()) {
//            skipFurtherOutput.set(true);
//          }
//        }
//        catch (VcsException e) {
//          exceptionConsumer.consume(e);
//        }
  }

  /**
   * Get name of the file in the last commit. If file was renamed, returns the previous name.
   *
   * @param project the contenxt project
   * @param path    the path to check
   * @return the name of file in the last commit or argument
   */
  public static FilePath getLastCommitName(final Project project, FilePath path) {
    final ChangeListManager changeManager = ChangeListManager.getInstance(project);
    final Change change = changeManager.getChange(path);
    if (change != null && change.getType() == Change.Type.MOVED) {
      BzrContentRevision r = (BzrContentRevision)change.getBeforeRevision();
      assert r != null : "Move change always have beforeRevision";
      path = r.getFile();
    }
    return path;
  }
}
